# *******************************************************************
# **  @file eisman_inc_info2rec.awk - convert info files to record format
# **
# **  Creation:     2018-12-27 dv
# **  Last update:  $Id: eisman_inc_info2rec.awk 63780 2019-03-15 22:23:43Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

@include "eisman_inc.awk"


# *******************************************************************
# **  convert an info file into the internal database format used
# **  by eisman.
# **  @param  recfile   record file used to write data to
# **  @param  infofile  info file used to read data from
# **  @param  name      name of package
# **  @param  version   version of package
# **  @param  url       url of package
# **  @return true if successful, false otherwise
# *******************************************************************

function info_to_record(recfile, infofile, name, version, url,   record)
{
    record["base_url"] = get_base_url(url);
    record["url"]      = url;

    if (info_to_record_read(infofile, record))
    {
        if (name != "")
        {
            record["name"] = name;
        }

        if (version != "")
        {
            record["version"] = version;
        }

        return info_to_record_write(recfile, record);
    }
    return false;
}


# *******************************************************************
# **  take a string as input and return it as a single line. Newline
# **  characters are converted into spaces, leading and trailing
# **  spaces are removed
# **  @param  line  is the input text string
# **  @return single line data
# *******************************************************************

function get_single_line_value(line)
{
    # place text on a single line
    gsub(/[\n\r\ŧ]/, " ", line);

    # remove leading spaces
    gsub(/^[ ]*/, "", line);

    # remove trailing spaces
    gsub(/[ ]*$/, "", line);
    return line;
}


# *******************************************************************
# **  read a package info file from xml format into an array supplied
# **  by the caller.
# **  @param  infofile  name of file to read data from
# **  @param  record    record set array to store data to
# **  @return true if successful, false if the data is invalid
# *******************************************************************

function info_to_record_read(infofile, record,   \
    encoding,
    tmpfile,
    line,
    tagname,
    text_count,
    textlist)
{
    tmpfile = infofile;

    record["require_count"]  = 0;
    record["link_count"]     = 0;
    record["replaces_count"] = 0;
    record["exclude_count"]  = 0;
    record["text_count"]     = 0;

    # check file encoding
    encoding = get_file_encoding(infofile);
    if ((encoding != "us_ascii") && (encoding != "unknown-8bit"))
    {
        tmpfile = "/tmp/encoded_file" PROCINFO["pid"];

        if (system("iconv -c -f " encoding " -t us-ascii " infofile " > " tmpfile) != 0)
        {
            remove(tmpfile);
            return false;
        }
    }

    # modify record separator
    OLD_RS = RS;
    RS="<[^>]+>"

    # initialise tag name
    tagname = "";

    # read input file line by line
    while ((getline line < tmpfile) > 0)
    {
        if (match(tagname, /<[^>\/]+>/))
        {
            # remove lesser bigger signs from tag name
            tagname = gensub(/[<>]/, "", "g", tagname);

            # remove leading and trailing newlines
            line = gensub(/^[\r\n]*/, "", "g", line);
            line = gensub(/[\r\n]*$/, "", "g", line);

            # if value is not empty, process it
            if (length(line) > 0)
            {
                switch(tagname)
                {
                # single line values
                case "name":
                case "short":
                case "version":
                case "author":
                case "status":
                case "section":
                case "system":
                case "space":
                case "sha1sum":
                case "sha256sum":
                    record[tagname] = get_single_line_value(line);
                    break;

                # special handling for dates
                case "date":
                    record[tagname] = gensub(/\//, "-", "g", get_single_line_value(line));
                    break;

                # multi line values
                case "require-lib":
                case "require-package":
                    line = get_single_line_value(line);
                    if (length(line) > 0)
                    {
                        record["require_count"]++;
                        record["require"][record["require_count"]] = line;
                    }
                    break;

                case "replaces":
                    line = get_single_line_value(line);
                    if (length(line) > 0)
                    {
                        record["replaces_count"]++;
                        record["replaces"][record["replaces_count"]] = line;
                    }
                    break;

                case "linked-package":
                    line = get_single_line_value(line);
                    if (length(line) > 0)
                    {
                        record["link_count"]++;
                        record["link"][record["link_count"]] = line;
                    }
                    break;

                case "exclude-package":
                    line = get_single_line_value(line);
                    if (length(line) > 0)
                    {
                        record["exclude_count"]++;
                        record["exclude"][record["exclude_count"]] = line;
                    }
                    break;

                # package description
                case "description":
                    text_count = split(line, textlist, "\n");
                    for (idx = 1; idx <= text_count; idx++)
                    {
                        record["text"][idx] = textlist[idx];
                    }
                    record["text_count"] = text_count;
                    break;
                }
            }
        }
        tagname = RT
    }
    close(tmpfile)

    # restore record separator
    RS = OLD_RS;

    # remove temp file if encoding hat do be changed
    if (tmpfile != infofile)
    {
        remove(tmpfile);
    }

    # if space is unknown, suggest it is the smallest size possible
    if (record["space"] == "")
    {
        record["space"] = "1";
    }

    # when no section is given, place it into "unknown"
    if (record["section"] == "")
    {
        record["section"] = "unknown";
    }

    # when no status is given, the package is taken as "unstable"
    if (record["status"] == "")
    {
        record["status"] = "unstable";
    }

    # when no short text is available, create one
    if (record["short"] == "")
    {
        record["short"] = record["name"] " package";
    }

    # check for minimum requirement
    if ((record["name"] != "") && (record["version"] != ""))
    {
        return true;
    }
    return false;
}


# *******************************************************************
# **  write record set read from an info file to the eisman info
# **  cache. The file already is written in database format.
# **  @param  recfile  name of record file to write to
# **  @param  record   record set of one single package
# **  @return true if successful, false if writing failed
# *******************************************************************

function info_to_record_write(recfile, record,   status, key, idx, dev_given)
{
    if (record["status"] == "stable")
    {
        status = "s";
    }
    else
    {
        status = "u";
    }

    if (!mkdir(eisman_path "/info/" record["name"] "/" record["version"]))
    {
        error("Failed to create info folder for \"" record["name"] "\"!");
        return false;
    }

    key = "[" record["section"] "]" \
          "[" record["name"]    "]" \
          "[" record["version"] "]" \
          "[" status "]"            \
          "[" record["base_url"] "]";

    print key "name       " record["name"]       > recfile;
    print key "version    " record["version"]   >> recfile;
    print key "short      " record["short"]     >> recfile;
    print key "date       " record["date"]      >> recfile;
    print key "author     " record["author"]    >> recfile;
    print key "section    " record["section"]   >> recfile;
    print key "space      " record["space"]     >> recfile;
    print key "sha1sum    " record["sha1sum"]   >> recfile;
    print key "sha256sum  " record["sha256sum"] >> recfile;
    print key "url        " record["url"]       >> recfile;
    print key "status     " record["status"]    >> recfile;

    for (idx = 1; idx <= record["replaces_count"]; idx++)
    {
        print key "replaces   " record["replaces"][idx]  >> recfile;
    }

    for (idx = 1; idx <= record["exclude_count"]; idx++)
    {
        print key "exclude    " record["exclude"][idx]  >> recfile;
    }

    for (idx = 1; idx <= record["require_count"]; idx++)
    {
        print key "require    " record["require"][idx]  >> recfile;
    }

    dev_given = false;
    for (idx = 1; idx <= record["link_count"]; idx++)
    {
        print key "link       " record["link"][idx]  >> recfile;

        if (index(record["link"][idx], record["name"] "-dev") == 1)
        {
            dev_given = true;
        }
    }

    # add implicit link to dev package needed for compatibility
    if (!dev_given)
    {
        if (!match(record["name"], /-dev$/) &&
            !match(record["name"], /-dev-static$/))
        {
            print key "link       " record["name"] "-dev " record["version"]  >> recfile;

            if (substr(record["name"], 1, 3) == "lib")
            {
                print key "link       " record["name"] "-dev-static " record["version"]  >> recfile;
            }
        }
    }

    for (idx = 1; idx <= record["text_count"]; idx++)
    {
        print key "text       " record["text"][idx]  >> recfile;
    }

    print key eisman_separator  >> recfile;

    close(recfile);

    return true;
}
