# *******************************************************************
# **  @file eisman_inc_version.awk - version handling functions
# **
# **  Creation:     2018-12-27 dv
# **  Last update:  $Id: eisman_inc_version.awk 65395 2019-05-20 20:01:43Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

@include "eisman_inc.awk"


# *******************************************************************
# **  split version into an array of version numbers. Normal versions
# **  contain three members, development versions may consist of up to
# **  five numbers. The number layout looks like this:
# **     ver_array[1] = major version
# **     ver_array[2] = minor version
# **     ver_array[3] = patch version
# **     ver_array[4] = development version postfix
# **     ver_array[5] = development version postfix timestamp (if given)
# **  @param  version  is a version number like "1.2.3"
# **  @param  ver_array  is the array to be filled with output data
# **  @return n/a
# *******************************************************************

function split_version(version, ver_array,   \
    ver_split,
    count,
    idx)
{
    # split off svn or cvs appendix into an array
    split(version, ver_split, /svn|cvs/);

    # now split first element of version array into pieces
    count = split(ver_split[1], ver_array, ".");

    # make sure, there are three version numbers
    for (idx = count + 1; idx < 3; idx++)
    {
        ver_array[idx] = "0";
    }

    # if there is a svn or cvs appendix
    if (ver_split[2] != "")
    {
        if (index(ver_split[2], ":"))
        {
            # this is for the svn_short tag MM-DDhh:mm
            # it comes with mktarball.sh -hometest
            ver_array[4] = strftime("%Y") "-" substr(ver_split[2], 1, 5);
            ver_array[5] = substr(ver_split[2], 6);
        }
        else
        {
            ver_array[4] = ver_split[2];
        }
    }
    else
    {
        ver_array[4] = "";
    }
}


# *******************************************************************
# **  This function performs version comparison:
# **  return value < 0 --> version1 is smaller (older) than version2
# **  return value > 0 --> version1 is bigger (newer) than version2
# **  return value = 0 --> version1 it the same as version2
# **  @param version1 is the version that is evaluated against version2
# **  @param version2 is the comparison value
# **  @return comparison result as described above
# *******************************************************************

function compare_versions(version1, version2, \
    ver_array1,
    ver_array2,
    idx)
{
    split_version(version1, ver_array1);
    split_version(version2, ver_array2);

    # compare major number
    if (ver_array1[1] != ver_array2[1])
    {
        return (ver_array1[1] + 0) - (ver_array2[1] + 0);
    }

    # compare minor number
    if (ver_array1[2] != ver_array2[2])
    {
        return (ver_array1[2] + 0) - (ver_array2[2] + 0);
    }

    # compare patch level
    if (ver_array1[3] != ver_array2[3])
    {
        return (ver_array1[3] + 0) - (ver_array2[3] + 0);
    }

    # if only one version number has a svn postfix, the version without
    # is the newer one
    if ((ver_array1[4] != "") && (ver_array2[4] == ""))
    {
        return -1;
    }
    else if ((ver_array1[4] == "") && (ver_array2[4] != ""))
    {
        return 1;
    }
    else if ((ver_array1[4] != "") && (ver_array2[4] != ""))
    {
        if (ver_array2[4] > ver_array1[4])
        {
            return -1;
        }
        else if (ver_array2[4] < ver_array1[4])
        {
            return 1;
        }
        return 0;
    }

    return 0;
}


# *******************************************************************
# **  Helper function for compare_versions. Returns "true" if
# **  version 1 is older than version 2
# **  @param version1 is the version that is evaluated against version2
# **  @param version2 is the comparison value
# **  @return boolean result
# *******************************************************************

function version_is_older(version1, version2)
{
    return compare_versions(version1, version2) < 0;
}


# *******************************************************************
# **  Helper function for compare_versions. Returns "true" if
# **  version 1 is newer than version 2
# **  @param version1 is the version that is evaluated against version2
# **  @param version2 is the comparison value
# **  @return boolean result
# *******************************************************************

function version_is_newer(version1, version2)
{
    return compare_versions(version1, version2) > 0;
}


# *******************************************************************
# **  Helper function for compare_versions that produces more readable
# **  result
# **  @param version1 is the version that is evaluated against version2
# **  @param version2 is the comparison value
# **  @return string result "old", "new", "installed"
# *******************************************************************

function evaluate_version(version1, version2,   \
    res)
{
    res = compare_versions(version1, version2);
    if (res < 0)
    {
        return "old";
    }
    else if (res > 0)
    {
        return "new";
    }
    else
    {
        return "installed";  # means equal
    }
}
