# *******************************************************************
# **  @file eisman_tree.awk - search for package dependencies
# **
# **  Creation:     2019-04-20 dv
# **  Last update:  $Id: eisman_tree.awk 65673 2019-05-31 15:53:34Z dv $
# **
# **  Copyright (c) 2019-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

@include "eisman_inc.awk"
@include "eisman_inc_query.awk"
@include "eisman_inc_version.awk"


# *******************************************************************
# ** management of result list
# *******************************************************************

# *******************************************************************
# **  search for a package named "name" within a packages array
# **  @param  name to search for
# **  @return index of entry in packages array or "0" if not found
# *******************************************************************

function find_package(name,   \
    idx)
{
    for (idx in result_packages)
    {
        if (result_packages[idx]["name"] == name)
        {
            return idx + 0;
        }
    }
    return 0;
}


# *******************************************************************
# **  add a package to the list of result packages
# **  @param  idx  is the element index of the package to add
# **  @return n/a
# *******************************************************************

function add_package(idx, level,  \
    pos)
{
    pos = ++num_result_packages;
    result_packages[pos]["name"]    = all_packages[idx]["name"];
    result_packages[pos]["version"] = all_packages[idx]["version"];
    result_packages[pos]["status"]  = all_packages[idx]["status"];
    result_packages[pos]["url"]     = all_packages[idx]["url"];
    result_packages[pos]["space"]   = all_packages[idx]["space"];
    result_packages[pos]["level"]   = level;

    if (result_packages[pos]["space"] == 0)
    {
        result_packages[pos]["space"] = 1;
    }
}


# *******************************************************************
# **  remove an element from a list of packages
# **  @param  idx  is the element index to remove
# **  @return n/a
# *******************************************************************

function remove_package(idx,  \
    pos)
{
    if (num_result_packages > 0)
    {
        pos = num_result_packages;
        while(idx < pos)
        {
            result_packages[idx]["name"]    = result_packages[idx + 1]["name"];
            result_packages[idx]["version"] = result_packages[idx + 1]["version"];
            result_packages[idx]["status"]  = result_packages[idx + 1]["status"];
            result_packages[idx]["url"]     = result_packages[idx + 1]["url"];
            result_packages[idx]["space"]   = result_packages[idx + 1]["space"];
            result_packages[idx]["level"]   = result_packages[idx + 1]["level"];
            idx++;
        }
        delete result_packages[pos];

        num_result_packages--;
    }
}


# *******************************************************************
# **  move single package to top of the list and set new index
# **  @param  idx    index of entry to move to top
# **  @param  level  new iteration level of the entry
# **  @return n/a
# *******************************************************************

function package_to_top(idx, level,   \
    name,
    version,
    status,
    space,
    url)
{
    if (idx < num_result_packages)
    {
        name    = result_packages[idx]["name"];
        version = result_packages[idx]["version"];
        status  = result_packages[idx]["status"];
        space   = result_packages[idx]["space"];
        url     = result_packages[idx]["url"];

        while(idx < num_result_packages)
        {
            result_packages[idx]["name"]    = result_packages[idx + 1]["name"];
            result_packages[idx]["version"] = result_packages[idx + 1]["version"];
            result_packages[idx]["status"]  = result_packages[idx + 1]["status"];
            result_packages[idx]["url"]     = result_packages[idx + 1]["url"];
            result_packages[idx]["space"]   = result_packages[idx + 1]["space"];
            result_packages[idx]["level"]   = result_packages[idx + 1]["level"];
            idx++;
        }

        result_packages[num_result_packages]["name"]    = name;
        result_packages[num_result_packages]["version"] = version;
        result_packages[num_result_packages]["status"]  = status;
        result_packages[num_result_packages]["space"]   = space;
        result_packages[num_result_packages]["url"]     = url;
        result_packages[num_result_packages]["level"]   = level;
    }
}


# *******************************************************************
# ** management of available packages
# *******************************************************************

# *******************************************************************
# **  search for a package named "name" within a packages array
# **  @param  name to search for
# **  @param  packages is the packages array to search in
# **  @return index of entry in packages array or "0" if not found
# *******************************************************************

function find_exact_match(name, version,  \
    idx)
{
    for (idx in all_packages)
    {
        if ((all_packages[idx]["name"]    == name) &&
            (all_packages[idx]["version"] == version) &&
            ((filter["status"] == "") || (filter["status"] == all_packages[idx]["status"])) &&
            ((filter["url"] == "")    || (index(all_packages[idx]["url"], filter["url"]) > 0)))
        {
            return idx + 0;
        }
    }
    return 0;
}


# *******************************************************************
# **  search for the best version of a package that is newer than
# **  a given version (newest update). In case of a requirement
# **  that can't be met by a already installed package, this is used
# **  to get the best version available.
# **  If -unstable switch has been specified the best version may have
# **  an unstable state, otherwise only stable packages are searched.
# **  @param  $1  package
# **  @param  $2  min. version (may be empty or 0.0.0 if not specified)
# **  @return (on stdout) best version found
# *******************************************************************

function find_best_match(name, version,   \
    idx,
    best_version,
    best_index,
    best_status)
{
    best_index = 0;

    if (version == "")
    {
        version = "0.0.0";
    }

    for (idx in all_packages)
    {
        if ((all_packages[idx]["name"] == name) &&
            ((filter["status"] == "") || (filter["status"] == all_packages[idx]["status"])) &&
            ((filter["url"] == "")    || (index(all_packages[idx]["url"], filter["url"]) > 0)) &&
            (evaluate_version(all_packages[idx]["version"], version) != "old"))
        {
            if (best_index == 0)
            {
                # if version has not been set yet, take the first match
                # even if it is unstable
                best_version = all_packages[idx]["version"];
                best_status  = all_packages[idx]["status"];
                best_index   = idx;
            }
            else
            {
                if ((best_status != "stable") && (all_packages[idx]["status"] == "stable"))
                {
                    # always prefer the stable package even if the unstable package is
                    # more recent
                    best_version = all_packages[idx]["version"];
                    best_status  = all_packages[idx]["status"];
                    best_index   = idx;
                }
                else if ((best_status != "stable") || (all_packages[idx]["status"] == "stable"))
                {
                    # only if the previous result is unstable or the newly found package is
                    # (also) stable, we proceed (never go from stable to unstable!):
                    # compare the version of both packages and take the one that is more recent
                    res = evaluate_version(all_packages[idx]["version"], best_version);
                    if (res == "new")
                    {
                        best_version = all_packages[idx]["version"];
                        best_status  = all_packages[idx]["status"];
                        best_index   = idx;
                    }
                }
            }
        }
    }
    return best_index + 0;
}


# *******************************************************************
# ** management of other packages
# *******************************************************************

# *******************************************************************
# **  search for a package named "name" in outdated packages
# **  @param  name to search for
# **  @return index of entry in packages array or "0" if not found
# *******************************************************************

function find_outdated_package(name)
{
    for (idx in outdated_packages)
    {
        if (outdated_packages[idx]["name"] == name)
        {
            return idx + 0;
        }
    }
    return 0;
}


# *******************************************************************
# **  search for a package named "name" in installed packages
# **  @param  name to search for
# **  @return index of entry in packages array or "0" if not found
# *******************************************************************

function find_installed_package(name)
{
    for (idx in installed_packages)
    {
        if (installed_packages[idx]["name"] == name)
        {
            return idx + 0;
        }
    }
    return 0;
}


# *******************************************************************
# **  dependency routines
# *******************************************************************

# *******************************************************************
# **  split a package reference (name/version) into an array
# **  @param  reference to be splitted
# **  @param  result of the operation
# **  @param  def_version is the default version to use, if nothing
# **          was specified
# **  @return true if successful
# *******************************************************************

function slit_package_ref(reference, result, def_version,  \
    count,
    token)
{
    count = split(reference, token, " ");
    if (count > 0)
    {
        if (count < 2)
        {
            result["name"]    = token[1];
            result["version"] = def_version;
        }
        else
        {
            result["name"]    = token[1];
            result["version"] = token[2];
        }
        return true;
    }
    return false;
}


# *******************************************************************
# **  check if the package issues exclusions
# **  @param  $1  name of package
# **  @param  $2  version of package
# **  @return n/a
# *******************************************************************

function check_exclusion(package_idx,   \
    exc_idx,
    idx)
{
    if (typeof(all_packages[package_idx]["exclude"]) == "array")
    {
        for (exc_idx in all_packages[package_idx]["exclude"])
        {
            # first check if excluded package is already installed
            idx = find_installed_package(all_packages[package_idx]["exclude"][exc_idx]);
            if (idx > 0)
            {
                print "error: processing package \""            \
                    all_packages[package_idx]["name"]    " ("   \
                    all_packages[package_idx]["version"] ")\":" \
                    > "/dev/stderr";

                print "error: exclusion conflict with \""                           \
                    all_packages[package_idx]["exclude"][exc_idx] "\" (installed)!" \
                    > "/dev/stderr";

                exit(1);
            }

            # check if package is in the list of packages that are going to
            # be installed
            idx = find_package(all_packages[package_idx]["exclude"][exc_idx]);
            if (idx > 0)
            {
                print "error: processing package \""            \
                    all_packages[package_idx]["name"]    " ("   \
                    all_packages[package_idx]["version"] ")\":" \
                    > "/dev/stderr";

                print "error: exclusion conflict with \""                        \
                    all_packages[package_idx]["exclude"][exc_idx] "\" (queued)!" \
                    > "/dev/stderr";

                exit(1);
            }
        }
    }
}


# *******************************************************************
# **  given is a certain requirement of a package. Check if it is already
# **  met by an installed package. If not search for the best
# **  available version and add the found package to the list
# **  of needed packages (if not already there).
# **  If the package has been added, it also may have requirements
# **  that have to be considered. Therefore the routine calls
# **  "find_requirements" or "find_linked_packages" and therefore enters
# **  a new recursion loop.
# **  Note: if a requirement cannot be resolved, an error is written
# **  to stdout and the whole script is terminated with exit code 1
# **  @param  $1  required package
# **  @param  $2  required version
# **  @param  $2  current iteration level
# **  @return n/a
# *******************************************************************

function check_and_add_requirement(package_idx, level,   \
    idx)
{
    idx = find_package(all_packages[package_idx]["name"]);
    if (idx == 0)
    {
        # package is new (currently not in result list)

        # check if package can be installed
        check_exclusion(package_idx);

        # add package to list
        add_package(package_idx, level);

        # find linked packages to this link (recursively)
        find_linked_packages(package_idx, level + 1);

        # find requirements (recursively)
        find_requirements(package_idx, level + 1);
    }
    else
    {
        # package already in the list, check if it's version is sufficient
        # for us. If not, remove the old package and (re)add the new one
        if (version_is_older(result_packages[idx]["version"],
            all_packages[package_idx]["version"]))
        {
            # listed package is old! This means we have to replace it with
            # the new one and have to deal with the new requirements.
            # This is a quite seldom and unusual situation and only happens, when
            # a stable package has to be replaced by an unstable one due to some
            # obscure combination of package requirements.
            # We only remove the single package and redo search for linked packages
            # and requirements. It may happen, that some requirements of the
            # original package remain in the list if the new version of the package
            # no longer requires them. But that is not really harmful! Therefore
            # we currently have to live with this behaviour

            # remove previous version from list
            remove_package(idx);

            # add package to list
            add_package(package_idx);

            # find linked packages to this link (recursively)
            find_linked_packages(package_idx, level + 1);

            # find requirements (recursively)
            find_requirements(package_idx, level + 1);
        }
        else
        {
            # listed package is sufficient for us! We only have to move it
            # to the top of the list and redo the requirement search to bring
            # up all required packages too. But we have no need to search
            # for linked packages at this point, since they already are in
            # the list.

            # move package to the top of the list
            package_to_top(idx, level);

            # find requirements (recursively)
            find_requirements(package_idx, level + 1);
        }
    }
}


# *******************************************************************
# **  get requirements of a package and check if they are already
# **  met by an installed package. If not search for the best
# **  available version and call "check_and_add_requirement"
# **  to add the new requirement to the list
# **  @param  $1  package
# **  @param  $2  version
# **  @return n/a
# *******************************************************************

function find_requirements(package_idx, level,    \
    req_idx,
    inst_idx,
    outdated_idx,
    entry,
    state,
    idx)
{
    if (typeof(all_packages[package_idx]["require"]) == "array")
    {
        for (req_idx in all_packages[package_idx]["require"])
        {
            if (slit_package_ref(all_packages[package_idx]["require"][req_idx],
                entry,
                "0.0.0"))
            {
                # special handling of kernel requirements, since it must be trans-
                # lated into it's installed variant
                if (entry["name"] == "eiskernel")
                {
                    entry["name"] = get_kernel_name();
                }

                # check if package is outdated
                outdated_idx = find_outdated_package(entry["name"]);
                if (outdated_idx > 0)
                {
                    if (outdated_packages[outdated_idx]["substitude"] != "dummy")
                    {
                        if (!slit_package_ref(outdated_packages[outdated_idx]["substitude"],
                            entry,
                            "0.0.0"))
                        {
                            # outdated entry is invalid
                            entry["name"] = "";
                        }
                    }
                    else
                    {
                        print "error: processing package \""            \
                            all_packages[package_idx]["name"]    " ("   \
                            all_packages[package_idx]["version"] ")\":" \
                            > "/dev/stderr";

                        print "error: requirement \""                  \
                            entry["name"] "\" is outdated!"            \
                            > "/dev/stderr";

                        print "error: " outdated_packages[outdated_idx]["hint"] \
                            > "/dev/stderr";

                        exit(1);
                    }
                }

                if (entry["name"] != "")
                {
                    # check if package is not installed or is too old to satisfy
                    # the new requirement
                    inst_idx = find_installed_package(entry["name"]);
                    if ((inst_idx == 0) ||
                        (version_is_older(installed_packages[inst_idx]["version"], entry["version"] "svn0")))
                    {
                        idx = find_best_match(entry["name"], entry["version"] "svn0");

                        # if the result is zero, then there is no package available
                        # that meets our needs
                        if (idx == 0)
                        {
                            print "error: processing package \""            \
                                all_packages[package_idx]["name"]    " ("   \
                                all_packages[package_idx]["version"] ")\":" \
                                > "/dev/stderr";

                            print "error: requirement \""                  \
                                entry["name"]    "\" >= \""                \
                                entry["version"] "\" not found!"           \
                                > "/dev/stderr";

                            exit(1);
                        }

                        # check if requirement is needed and recursively add the package
                        # and all of it's requirements
                        check_and_add_requirement(idx, level);
                    }
                }
            }
        }
    }
}


# *******************************************************************
# **  find all packages that are specified with the linked-package
# **  tag. This packages are added to the requirement list if
# **  they are installed on the machine and the specified version
# **  is newer than the installed on
# **  @param  $1  idx  the package index
# **  @param  $2  version
# **  @param  $3  level
# **  @return n/a
# *******************************************************************

function find_linked_packages(package_idx, level,   \
    link_idx,
    inst_idx,
    outdated_idx,
    entry,
    state,
    idx)
{
    if (typeof(all_packages[package_idx]["link"]) == "array")
    {
        for (link_idx in all_packages[package_idx]["link"])
        {
            if (slit_package_ref(all_packages[package_idx]["link"][link_idx],
                entry,
                all_packages[package_idx]["version"]))
            {
                # special handling of kernel requirements, since it must be trans-
                # lated into it's installed variant
                if (entry["name"] == "eiskernel")
                {
                    entry["name"] = get_kernel_name();
                }

                # if the package in question is installed and it's version is older
                # than the requested one, we need to update the package!
                inst_idx = find_installed_package(entry["name"]);
                if (inst_idx > 0)
                {
                    if (version_is_newer(entry["version"] "svn0", installed_packages[inst_idx]["version"]))
                    {
                        # if the package that is linked to is in outdated packages but is
                        # installed under it's original name, then it obviously has been
                        # renamed or has been replaced by an other package
                        outdated_idx = find_outdated_package(entry["name"]);
                        if (outdated_idx > 0)
                        {
                            if (outdated_packages[outdated_idx]["substitude"] != "dummy")
                            {
                                if (!slit_package_ref(outdated_packages[outdated_idx]["substitude"],
                                    entry,
                                    "0.0.0"))
                                {
                                    # outdated entry is invalid
                                    entry["name"] = "";
                                }
                            }
                            else
                            {
                                # if the successor is dummy, there is no replacement available
                                # in this case we ignore the linkage by setting the package
                                # name to zero
                                entry["name"] = "";
                            }
                        }

                        if (entry["name"] != "")
                        {
                            idx = find_best_match(entry["name"], entry["version"] "svn0");

                            # if the result is zero, then there is no package available
                            # that meets our needs
                            if (idx == 0)
                            {
                                print "error: processing package \""            \
                                    all_packages[package_idx]["name"]    " ("   \
                                    all_packages[package_idx]["version"] ")\":" \
                                    > "/dev/stderr";

                                print "error: link requirement \""   \
                                    entry["name"]    "\" >= \""      \
                                    entry["version"] "\" not found!" \
                                    > "/dev/stderr";

                                exit(1);
                            }

                            # check if requirement is needed and recursively add the package
                            # and all of it's requirements
                            check_and_add_requirement(idx, level);
                        }
                    }
                }
            }
        }
    }
}


# *******************************************************************
# **  do complete package processing (search for requirements).
# **  in case of an error, it is output to stderr and the script is
# **  terminated with exit code 1
# **  @param  $1  package
# **  @param  $2  version
# **  @return n/a
# *******************************************************************

function process_root_package(name, version,   \
    package_idx,
    exclude_idx)
{
    # check if package exists
    package_idx = find_exact_match(name, version);
    if (package_idx == 0)
    {
        print "error: package \"" name "\" not found!" > "/dev/stderr";
        exit(1);
    }

    # check if package can be installed
    check_exclusion(package_idx);

    # add the root package
    add_package(package_idx, 0);

    # when the root package has been found, any other required package
    # may be from any source. Therefore reset URL filter
    filter["url"] = "";

    # even if the global search mode tells us to prefer only stable
    # packages, any required package may also be unstable. Therefore
    # reset state filter
    filter["status"] = "";

    # find linked packages to this package.
    find_linked_packages(package_idx, 1);

    # search for requirements. From here on, no further URL or status
    # filtering is done.
    find_requirements(package_idx, 1);
}


# *******************************************************************
# **  start package processing
# **  @param  $1  package
# **  @param  $2  version (optional)
# **  @return n/a
# *******************************************************************

function process_single_package(name, version,    \
    idx)
{
    # check if package is outdated. In case of a root package, we
    # issue an error and exit. In case of requirements we try to
    # use the substitude found in outdated db.
    idx = find_outdated_package(name);
    if (idx > 0)
    {
        print "error: processing package \"" name "\":"  > "/dev/stderr";
        print "error: package is outdated!"              > "/dev/stderr";
        print "error: " outdated_packages[idx]["hint"]   > "/dev/stderr";
        exit(1);
    }

    # if version is not specified, search for best available package
    # if not found output error and exit script
    if (version == "")
    {
        idx = find_best_match(name, "0.0.0");
        if (idx == 0)
        {
            print "error: processing package \"" name "\":" > "/dev/stderr";
            print "error: package not found!"               > "/dev/stderr";
            exit(1);
        }
        else
        {
            version = all_packages[idx]["version"];
        }
    }

    # check if package is already in the tree (propably has been a
    # requirement found in a previous run of 'process_single_package')
    idx = find_package(name);
    if (idx > 0)
    {
        # package already in the list, check if it's version is sufficient
        # for us. If not, remove the old package and (re)add the new one
        if (evaluate_version(result_packages[idx]["version"], version) == "old")
        {
            # remove previous version from list
            remove_package(idx);

            # process the needed package (root of requirement tree)
            process_root_package(name, version);
        }
    }
    else
    {
        # process the needed package (root of requirement tree)
        process_root_package(name, version);
    }
}


# *******************************************************************
# entry function of awk script
# accepts global parameters (passed with -v"
#   packages = list of space separated package names (form <name>[=<version>])
#   unstable = "yes"|"no"  accept unstable packages
#   url      = <url> to be used for package search
# *******************************************************************

BEGIN {
    installed_db = eisman_path "/installed.db";
    packages_db  = eisman_path "/packages.db";
    outdated_db  = eisman_path "/outdated.db";

    num_result_packages = 0;

    # query packages intalled on this system
    query_package_db( \
        "",
        "",
        "",
        "",
        "",
        "",
        "name|version|status",
        installed_db,
        installed_packages);

    # query available packages
    query_package_db( \
        "",
        "",
        "",
        "",
        "",
        "",
        "name|version|status|require|link|exclude|url|space",
        packages_db,
        all_packages);

    # query outdated packages
    query_package_db( \
        "",
        "",
        "",
        "",
        "",
        "",
        "name|version|hint|substitude",
        outdated_db,
        outdated_packages);

    # prepare package list from packages
    split(packages, package_list, " ");
    for (i in package_list)
    {
        gsub(/=/, " ", package_list[i]);

        if (slit_package_ref(package_list[i], result, ""))
        {
            # setup status filter
            if (unstable == "yes")
            {
                filter["status"] = "";
            }
            else
            {
                filter["status"] = "stable";
            }

            # setup url filter
            if (url != "")
            {
                filter["url"] = url;
            }

            process_single_package(result["name"], result["version"]);
        }
    }

    if (num_result_packages > 0)
    {
        total = 0;

        for (i = num_result_packages; i > 0; i--)
        {
            printf("%-8s %-8s %-17s %s\n",
                result_packages[i]["version"],
                result_packages[i]["status"],
                result_packages[i]["name"],
                gensub(/\/[^/]*$/, "", "g", result_packages[i]["url"]));

            total += result_packages[i]["space"];
        }

        print "space = " total;
    }
}

