# *******************************************************************
# **  @file eisman_update.awk - update local package database
# **
# **  Creation:     2018-12-27 dv
# **  Last update:  $Id: eisman_update.awk 65952 2019-06-10 09:04:02Z dv $
# **
# **  Copyright (c) 2013-2019the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

@include "eisman_inc.awk"
@include "eisman_inc_query.awk"
@include "eisman_inc_info2rec.awk"


# *******************************************************************
# entry function of awk script
# *******************************************************************

BEGIN {
    if (progress == "")
    {
        progress = yes;
    }

    if (quiet == "")
    {
        quiet = yes;
    }

    indexfile = "/tmp/index" PROCINFO["pid"] ".txt";
    database  = eisman_path "/packages.db";

    # fetch index and prepare a filtered version of the fetched
    # index file that only contains data relevant for us
    if (prepare_index(indexfile))
    {
        # read contents of info file into the global array "package"
        # and return the number of records found
        records = read_records(indexfile);
        if (records > 0)
        {
            # prepare an empty temporary file that will become the
            # new packages database
            remove(database ".tmp");
            touch(database ".tmp");

            # iterate over all records and create the packages database
            if (quiet != "yes")
            {
                print_info("(Re)building package database ...");
                process_records(records, database ".tmp");
            }
            else
            {
                process_records(records, database ".tmp");
            }

            # rename the temporary version of the new packages database
            # into it's final name
            close(database ".tmp");
            remove(database);
            rename(database ".tmp", database);

            # use the new database version to update the database of
            # the outdated packages (by extracting the "replaces" info)
            if (quiet != "yes")
            {
                print_info("(Re)building outdated database ...");
                update_outdated_packages();
            }
            else
            {
                update_outdated_packages();
            }

            # clean up files that are older than 90 days (meaning they
            # never appeared within an index file for this time)
            if (quiet != "yes")
            {
                print_info("Cleaning up info cache ...");
                clean_up_info_cache();
            }
            else
            {
                clean_up_info_cache();
            }
        }

        # remove our filtered version of the index file
        remove(indexfile);
    }
    else
    {
        error("Unable to fetch index!");
        exit 1;
    }
}


# *******************************************************************
# **  read all records from the index files and store them into the
# **  global array 'package'
# **  @param  idx_file  full name of (temporary) index file
# **  @return number of records read from file
# *******************************************************************

function read_records(idx_file,    \
    line,
    fields,
    num,
    row)
{
    row = 0;
    while ((getline line < idx_file) > 0)
    {
        num = split(line, fields);
        if (num > 4)
        {
            if (substr(fields[1], 1, 1) != "#")
            {
                row++;
                package[row]["name"]     = fields[1];
                package[row]["version"]  = fields[2];
                package[row]["state"]    = fields[3];
                package[row]["info_url"] = fields[4];
                package[row]["pack_url"] = substr(fields[4], 1, length(fields[4]) - 5);

                if (num < 5)
                {
                    package[row]["system"] = "eisfair-1";
                }
                else
                {
                    package[row]["system"]  = fields[5];
                }
            }
            else if (fields[2] == "required_base:")
            {
                print "Required base: " fields[3];
            }
        }
    }
    close(idx_file)

    return row;
}


# *******************************************************************
# **  process all records stored in global array 'package'.
# **  - Fetch info file from server if not already there
# **  - Add info file to package database
# **  @param  total_count  total count of records stored in array
# **  @param  database     name of database file to store data to
# *******************************************************************

function process_records(total_count, database,   \
    row,
    recfile,
    filespec,
    num,
    filename,
    touchfiles)
{
    show_progress(0);

    for (row = 1; row <= total_count; row++)
    {
        # split url into path and filename
        num = split(package[row]["pack_url"], filespec, "/");
        if (num > 0)
        {
            filename = filespec[num];
            recfile  = eisman_path      "/info/" \
                package[row]["name"]    "/"      \
                package[row]["version"] "/"      \
                filename ".rec";

            if (!file_exists(recfile))
            {
                printf("=> Fetching " package[row]["name"] " (" package[row]["version"] ") ...")

                # download new info file
                if (download(package[row]["info_url"],
                    "/tmp/tmpinfofile" PROCINFO["pid"]))
                {
                    # convert the info file into the record format and
                    # store it to the info cache
                    if (info_to_record(recfile,
                        "/tmp/tmpinfofile" PROCINFO["pid"],
                        package[row]["name"],
                        package[row]["version"],
                        package[row]["pack_url"]))
                    {
                        print_ok();

                        # conversion succeeded, now add the new file to
                        # to the packages database
                        add_to_database(row, recfile, database);
                    }
                    else
                    {
                        print_fail();
                    }
                }
                else
                {
                    print_fail();
                }

                # remove remainder of previous download
                remove("/tmp/tmpinfofile" PROCINFO["pid"]);

                show_progress(row * 100 / total_count);
            }
            else
            {
                # the record file already exists within the info cache
                # but we will touch it, to change it's time and date
                # to now
                touchfiles = touchfiles " " recfile;

                add_to_database(row, recfile, database);

                if (row % 25 == 0)
                {
                    show_progress(row * 100 / total_count);
                }
            }
        }
    }

    # we use one single call to touch, to change all files at once
    touch(touchfiles);

    show_progress(100);
    end_progress();
}


# *******************************************************************
# **  update database of outdated packages. For this purpose, the
# **  new version of the packages database is searched for packages
# **  having "replaces" info lines. This packages are further processed
# **  to generate the oudated packages database
# *******************************************************************

function update_outdated_packages(    \
    packages_db,
    outdated_db,
    result,
    count,
    idx,
    status,
    values,
    replaces_idx)
{
    packages_db = eisman_path "/packages.db";
    outdated_db = eisman_path "/outdated.db";

    # first copy fixed part of outdated db
    copy(eisman_path "/outdated.inc.db", outdated_db);

    # now query for all packages
    count = query_package_db( \
        "",
        "",
        "",
        "",
        "",
        "",
        "name|version|status|replaces",
        packages_db,
        result);

    show_progress(0);

    # search result set for packages with "replaces"-info
    for (idx = 1; idx <= count; idx++)
    {
        replaces_idx = 1;
        while (result[idx]["replaces"][replaces_idx] != "")
        {
            # split "replaces"-info into name and version
            split(result[idx]["replaces"][replaces_idx], values);
            if (values[2] == "")
            {
                values[2] = "0.0.0";
            }

            # make short status
            status = "u";
            if (result[idx]["status"] == "stable")
            {
                status = "s";
            }

            # create key common for all output lines
            key = "[outdated][" \
                  values[1] "][" \
                  result[idx]["version"]  "][" \
                  status  "][]";

            print key "name       " values[1]   >> outdated_db;
            print key "version    " values[2]   >> outdated_db;
            print key "hint       replaced by or included in " result[idx]["name"]    >> outdated_db;
            print key "substitude " result[idx]["name"] " "    result[idx]["version"] >> outdated_db;
            print key eisman_separator          >> outdated_db;

            replaces_idx++;
        }

        if (idx % 25 == 0)
        {
            show_progress(idx * 100 / count);
        }
    }

    show_progress(100);
    end_progress();

    close(outdated_db);
}


# *******************************************************************
# **  remove files from info cache older than 90 days
# *******************************************************************

function clean_up_info_cache(    \
    files,
    path_list,
    num_paths,
    num_files,
    idx)
{
    show_progress(0);
    num_files = 0;

    num_paths = split( \
        "a b c d e f g h i j k l m n o p q r s t u v w x y z ",
        path_list);

    # since find takes some time to complete, it's call is splitted
    # into 26 separated calls, so that a progress indicator can be shown
    for (idx in path_list)
    {
        num_files += find_files( \
            files,
            eisman_path "/info/" path_list[idx] "*",
            "-mtime +90 -name \"*.rec\"");

        show_progress((idx * 100) / num_paths);
    }

    # we expect num_files to be a very low number of files. Therefore
    # deletion of these files takes place when progress indicator already 
    # shows 100%
    if (num_files > 0)
    {
        for (idx = 1; idx <= num_files; idx++)
        {
            remove(files[idx]);
            rmdir(dirname(files[idx]));
        }
    }

    show_progress(100);
    end_progress();
}


# *******************************************************************
# **  add a record file (from info cache) to the database of available
# **  packages
# **  @param  row       row index
# **  @param  recfile   name of record file to add
# **  @param  database  name of database to add record file to
# *******************************************************************

function add_to_database(row, recfile, database,   \
    pack_url,
    base_url,
    line,
    idx,
    version,
    filter_pos)
{
    pack_url = package[row]["pack_url"];
    base_url = get_base_url(pack_url);

    # fetch version string from record file
    while ((getline line < recfile) > 0)
    {
        idx = match(line, /]version    /);
        if (idx > 0)
        {
            version = substr(line, idx + 12, length(line));
            break;
        }
    }
    close(recfile)

    # if version matches, copy to database
    if (version = package[row]["version"])
    {
        while ((getline line < recfile) > 0)
        {
            line = gensub(/\]url        .*/, "]url        " pack_url, "g", line);
            line = gensub(/\r/, "", "g", line);
            line = gensub(/\[[^][]*\]/, "[" base_url "]", 5 , line);

            # write input line to database
            print line >> database;

            # if version line has been written add extra virtual line "record"
            idx = match(line, /]version    /);
            if (idx > 0)
            {
                print substr(line, 1, idx) "record     " row >> database;
            }
        }
        close(recfile)
    }
    else
    {
        print_warn("warning: version mismatch! " \
            package[row]["name"]    " ("         \
            package[row]["version"] ") ignored!");
    }
}


# *******************************************************************
# **  fetch and prepare remote server's file index
# **  @param  idx_file    name of index file
# **  @return true if successful, false otherwise
# *******************************************************************

function prepare_index(idx_file)
{
    # if this is the first time the script is being executed, we
    # propably have to create the info cache folder first
    if (!mkdir(eisman_path "/info"))
    {
        error("Failed to create \"" eisman_path "/info" "\"!");
        return false;
    }

    # now the script "get-index" is used to download and combine
    # index files.
    if (exec_command("/var/install/bin/get-index --clear") != 0)
    {
        error("Failed to fetch index from server!");
        return false;
    }

    # now filter out all lines with packages marked to belong to a
    # different system
    if (exec_command( \
           "egrep '^.* ([^;]*[;])*(" eisman_noarch "|" eisman_system ")([;][^;]*)*$' " \
           "/tmp/index.txt " \
           "> " idx_file) != 0)
    {
        error("Failed to prepare index file!");
        return false;
    }

    return true;
}


# *******************************************************************
# **  show progress of update
# **  @param  pos  percentage progress value
# *******************************************************************

function show_progress(pos)
{
    if (progress == "yes")
    {
        printf("\033[7D[%3d%%] completed!\033[10D", pos);
    }
}


# *******************************************************************
# **  end output of progress with a newline 
# *******************************************************************

function end_progress()
{
    if (progress == "yes")
    {
        print;
    }
}


# *******************************************************************
# **  download file from remote server using wget
# **  @param  url       url to load
# **  @param  filename  output file name
# **  @return true if successful, false otherwise
# *******************************************************************

function download(url, filename,    \
    exitcode)
{
    exitcode = exec_command("/usr/bin/wget.sh -q -O " filename " " url);
    return (exitcode == 0);
}
