# *******************************************************************
# **  @file eisman_upgrade.awk - search and install upgradable packages
# **
# **  Creation:     2019-04-16 dv
# **  Last update:  $Id: eisman_upgrade.awk 64642 2019-04-22 07:52:13Z hbfl $
# **
# **  Copyright (c) 2019-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

@include "eisman_inc.awk"
@include "eisman_inc_query.awk"
@include "eisman_inc_version.awk"


# *******************************************************************
# **  convert list of packages into a string of concatenated package
# **  names separated by '|'
# **  @param  packages  is a list of packages (array)
# **  @return string of package names
# *******************************************************************

function get_package_list(packages,   \
   result,
   idx)
{
    result = "";
    for (idx in packages)
    {
        if (result == "")
        {
            result = packages[idx]["name"];
        }
        else
        {
            result = result "|" packages[idx]["name"];
        }
    }
    return result;
}


# *******************************************************************
# **  interate over all available packages to look for suitable updates
# **  for a specific package. If an update was found, it's index is
# **  returned, otherwise the return value is 0
# **  @param  package  the package to search an update for
# **  @param  packages is a list of available packages (array)
# **  @return index of suitable update package or 0 if nothing could
# **          be found
# *******************************************************************

function search_for_updates(package, packages,    \
    idx,
    best_version,
    best_index)
{
    best_index   = 0;
    best_version = package["version"];

    for (idx in packages)
    {
        if ((package["name"] == packages[idx]["name"]) &&
            (version_is_newer(packages[idx]["version"], best_version)))
        {
            best_version = packages[idx]["version"];
            best_index = idx;
        }
    }

    return best_index;
}


# *******************************************************************
# **  search for upgradable packages
# **  @param  upgrade_list
# **  @param  state_filter
# **  @param  exclusions is a list of packages excluded from update
# **  @return number of upgrades found
# *******************************************************************

function find_upgradables(upgrade_list, state_filter, exclusions,   \
    filename,
    idx,
    num_upgrades)
{
    installed_db = eisman_path "/installed.db";
    packages_db  = eisman_path "/packages.db";
    num_upgrades = 0;

    # append a leading and a terminating comma as field separator to
    # exclusions
    exclusions = "," exclusions ",";

    # query packages intalled on this system
    icount = query_package_db( \
        "",
        "",
        "",
        "",
        "",
        "",
        "name|version",
        installed_db,
        installed_packages);

    if (icount > 0)
    {
        # query all installed package names from the available packages
        # database
        pcount = query_package_db( \
            "",
            "",
            get_package_list(installed_packages),
            "",
            state_filter,
            "",
            "name|short|status|date|version|url",
            packages_db,
            available_packages);

        if (pcount > 0)
        {
            for (idx in installed_packages)
            {
                # check if package is part of the exclusions list
                pos = index(exclusions, "," installed_packages[idx]["name"] ",");
                if (pos == 0)
                {
                    # package is not excluded, search for updates
                    best_idx = search_for_updates(installed_packages[idx], available_packages);
                    if (best_idx)
                    {
                        num_upgrades++;

                        upgrade_list[num_upgrades]["name"]    = available_packages[best_idx]["name"];
                        upgrade_list[num_upgrades]["version"] = available_packages[best_idx]["version"];
                        upgrade_list[num_upgrades]["short"]   = available_packages[best_idx]["short"];
                        upgrade_list[num_upgrades]["date"]    = available_packages[best_idx]["date"];
                        upgrade_list[num_upgrades]["url"]     = available_packages[best_idx]["url"];
                        upgrade_list[num_upgrades]["status"]  = available_packages[best_idx]["status"];
                    }
                }
            }
        }
    }
    return num_upgrades;
}


# *******************************************************************
# entry function of awk script
# accepts global parameters (passed with -v"
#   progress="yes"|"no"  show wait animation (default = "yes")
#   unstable="yes"|"no"  accept unstable updates too (default = "no")
#   action  ="list"|"install"  action to perform (default = "")
#   exclude = <list of package names to exclude, separated by comma>
# *******************************************************************

BEGIN {
    if (unstable == "yes")
    {
        num = find_upgradables(upgrade_list, ".*", exclude);
    }
    else
    {
        num = find_upgradables(upgrade_list, "s", exclude);
    }

    if (action == "install")
    {
        if (num == 0)
        {
            print "-------------------------------------";
            print "Your system is up to date!";
            print "-------------------------------------";
        }
        else
        {
            # build sequential list of packages and versions
            for (rec_idx in upgrade_list)
            {
                package_list = package_list " " \
                    upgrade_list[rec_idx]["name"] "=" upgrade_list[rec_idx]["version"];
            }

            # use optional argument --unstable
            if (unstable == "yes")
            {
                arguments = arguments " --unstable";
            }

            # disable wait animation
            if (progress == "no")
            {
                arguments = arguments " --no-wait-anim";
            }

            exec_command(eisman_bin_path "/eisman_install.sh" arguments " " package_list);
        }
    }
    else if (action == "list")
    {
        list_packages(upgrade_list);
    }

    exit (num > 0 ? 0 : 1);
}
