#!/bin/bash
# *******************************************************************
# **  @file eisman_query.sh - query package database
# **
# **  Creation:     2013-01-13 dv
# **  Last update:  $Id: eisman_query.sh 64414 2019-04-16 19:13:45Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

. /var/install/include/eislib
. /var/install/include/eismanlib

gawk=/usr/bin/gawk

# *******************************************************************
# **  show help on arguments and possible options
# **  @param  n/a
# **  @return n/a
# *******************************************************************

show_help()
{
cat <<EOF

Query package database

usage:
    eisman query [OPTIONS] [PACKAGE] [VERSION]

options:
    --help                      show this help screen
    --excl-sections <sec-list>  comma separated list of sections
    --sections      <sec-list>  comma separated list of sections
    --fields        <fld-list>  comma separated list of fields
    --unstable                  show unstable packages alongside of
                                stable ones
    --installed                 query installed packages database
                                note: implicitly sets --unstable too
    --outdated                  query outdated packages database
    --output=<file>             redirect result into a file
    --url           <url>       only search packages from this source
    --quiet                     perform quiet operation
    --no-tty                    don't use escape sequences for cursor
                                movement and color output.
possible values for fields:
    name     package name
    version  package version
    short    short package description
    date     date of creation
    author   author of the package
    section  section of the package
    space    disk space needed in MB
    url      URL of the package
    status   status of the package
    require  required packages (multiple results possible)
    link     linked packages (that need update if installed)
    text     descriptive text (multiple results possible)

PACKAGE:
    a package name that may contain wildcards like
    * -> matches multiple characters
    ? -> matches exactly one character
    example:  *sql  matches postgresql, mysql,...
              ??sql matches mysql, ...

VERSION:
    a version string that may contain wildcards too

Note:
    By default, the source of input is the database of the available
    packages. This can be changed by either using:
    --installed   to select the installed packages database or
    --outdated    to select the outdated packages database.
    These options are mutual exclusive!

EOF
}


# *******************************************************************
# **  main program
# *******************************************************************
main()
{
    # initialize defaults
    local quiet="no"
    local state="s"
    local sections
    local exclsections
    local fields
    local packages
    local version
    local filterurl
    local database
    local outfile

    # check if output device is a pipe
    if [ -t 1 ]
    then
        local is_tty="yes"
    else
        local is_tty="no"
    fi

    # read option switches
    while [ "${1:0:2}" = "--" ]
    do
        case ${1} in
        --help)
            shift
            show_help
            exit 0
            ;;
        --sections)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                sections="${1}"
                shift
            else
                echo 'error: missing argument for --sections!' >&2
                show_help
                exit 1
            fi
            ;;
        --sections=*)
            sections="${1:11}"
            shift
            ;;
        --excl-sections)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                exclsections="${1}"
                shift
            else
                echo 'error: missing argument for --excl-sections!' >&2
                show_help
                exit 1
            fi
            ;;
        --excl-sections=*)
            exclsections="${1:16}"
            shift
            ;;
        --fields)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                fields="${1}"
                shift
            else
                echo 'error: missing argument for --fields!' >&2
                show_help
                exit 1
            fi
            ;;
        --fields=*)
            fields="${1:9}"
            shift
            ;;
        --unstable)
            state=".*"
            shift
            ;;
        --url)
            shift
            filterurl="$(echo "${1}" | sed 's#http[s]*://#http(s)?://#').*"
            shift
            ;;
        --url=*)
            filterurl="$(echo "${1:6}" | sed 's#http[s]*://#http(s)?://#').*"
            shift
            ;;
        --output)
            shift
            outfile="${1}"
            shift
            ;;
        --output=*)
            outfile="${1:9}"
            shift
            ;;
        --installed)
            database="${EISMAN_PATH}/installed.db"
            state=".*"
            shift
            ;;
        --outdated)
            database="${EISMAN_PATH}/outdated.db"
            shift
            ;;
        --quiet)
            quiet="yes"
            shift
            ;;
        --no-tty)
            is_tty="no"
            shift
            ;;
        *)
            echo "error: invalid switch \"$1\"!" >&2
            show_help
            exit 1
            ;;
        esac
    done

    if [ -n "${1}" ]
    then
        packages="${1}"
        shift
    fi

    if [ -n "${1}" ]
    then
        version="${1}"
        shift
    fi

    # if quiet operation is requested, redirect output
    if [ "${quiet}" = "yes" ] && [ -z "${outfile}" ]
    then
        outfile="/dev/null"
    fi

    # set include path for awk scripts
    export AWKPATH="${EISMAN_BIN_PATH}/awk"

    if [ -z "${outfile}" ]
    then
        ${gawk} -v incl_sections="${sections}"      \
                -v excl_sections="${exclsections}"  \
                -v packages="${packages}"             \
                -v version="${version}"             \
                -v state="${state}"                 \
                -v url_filter="${filterurl}"        \
                -v fields="${fields}"               \
                -v database="${database}"           \
                -v is_tty="${is_tty}"               \
                -v color="${is_tty}"                \
                -f "${EISMAN_BIN_PATH}/awk/eisman_query.awk"
    else
        ${gawk} -v incl_sections="${sections}"      \
                -v excl_sections="${exclsections}"  \
                -v packages="${packages}"           \
                -v version="${version}"             \
                -v state="${state}"                 \
                -v url_filter="${filterurl}"        \
                -v fields="${fields}"               \
                -v database="${database}"           \
                -v is_tty="no"                      \
                -v color="no"                       \
                -f "${EISMAN_BIN_PATH}/awk/eisman_query.awk" \
                > "${outfile}"
    fi

    exit ${?}
}

main "${@}"
