#!/bin/bash
# *******************************************************************
# **  @file eisman_search.sh - fuzzy search in package database
# **
# **  Creation:     2013-01-13 dv
# **  Last update:  $Id: eisman_search.sh 64414 2019-04-16 19:13:45Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

. /var/install/include/eislib
. /var/install/include/eismanlib

gawk=/usr/bin/gawk

# *******************************************************************
# **  show help on arguments and possible options
# **  @param  n/a
# **  @return n/a
# *******************************************************************

show_help()
{
cat <<EOF

Fuzzy search package database

usage:
    eisman search [OPTIONS] [SEARCH WORD]

options:
    --help                      show this help screen
    --excl-sections <sec-list>  comma separated list of sections
    --sections      <sec-list>  comma separated list of sections
    --fields        <fld-list>  comma separated list of fields
    --unstable                  show unstable packages alongside of
                                stable ones
    --installed                 query installed packages database
                                note: implicitly sets --unstable too
    --outdated                  query outdated packages database
    --output=<file>             redirect result into a file
    --url           <url>       only search packages from this source
    --quiet                     perform quiet operation
    --no-tty                    don't use escape sequences for cursor
                                movement and color output.
possible values for fields:
    name     package name
    version  package version
    short    short package description
    date     date of creation
    author   author of the package
    section  section of the package
    space    disk space needed in MB
    url      URL of the package
    status   status of the package
    require  required packages (multiple results possible)
    link     linked packages (that need update if installed)
    text     descriptive text (multiple results possible)

SEARCH WORD:
    a character string that may exist as substring withing a package
    name

EOF
}


# *******************************************************************
# **  main program
# *******************************************************************
main()
{
    # initialize defaults
    local quiet="no"
    local state="s"
    local sections
    local exclsections
    local fields
    local package
    local version
    local filterurl
    local database
    local outfile

    # check if output device is a pipe
    if [ -t 1 ]
    then
        local is_tty="yes"
    else
        local is_tty="no"
    fi

    # read option switches
    while [ "${1:0:2}" = "--" ]
    do
        case ${1} in
        --help)
            shift
            show_help
            exit 0
            ;;
        --sections)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                sections="${1}"
                shift
            else
                echo 'error: missing argument for --sections!' >&2
                show_help
                exit 1
            fi
            ;;
        --sections=*)
            sections="${1:11}"
            shift
            ;;
        --excl-sections)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                exclsections="${1}"
                shift
            else
                echo 'error: missing argument for --excl-sections!' >&2
                show_help
                exit 1
            fi
            ;;
        --excl-sections=*)
            exclsections="${1:16}"
            shift
            ;;
        --fields)
            shift
            if [ "${1}" != "" -a "${1:0:2}" != "--" ]
            then
                fields="${1}"
                shift
            else
                echo 'error: missing argument for --fields!' >&2
                show_help
                exit 1
            fi
            ;;
        --fields=*)
            fields="${1:9}"
            shift
            ;;
        --unstable)
            state=".*"
            shift
            ;;
        --url)
            shift
            filterurl="$(echo "${1}" | sed 's#http[s]*://#http(s)?://#').*"
            shift
            ;;
        --url=*)
            filterurl="$(echo "${1:6}" | sed 's#http[s]*://#http(s)?://#').*"
            shift
            ;;
        --output)
            shift
            outfile="${1}"
            shift
            ;;
        --output=*)
            outfile="${1:9}"
            shift
            ;;
        --installed)
            database="${EISMAN_PATH}/installed.db"
            state=".*"
            shift
            ;;
        --outdated)
            database="${EISMAN_PATH}/outdated.db"
            shift
            ;;
        --quiet)
            quiet="yes"
            shift
            ;;
        --no-tty)
            is_tty="no"
            shift
            ;;
        *)
            echo "error: invalid switch \"$1\"!" >&2
            show_help
            exit 1
            ;;
        esac
    done

    if [ -n "${1}" ]
    then
        package="${1}"
        shift
    fi

    local pattern1="${package}"
    local pattern2="$(echo "${package}" | sed 's#-#_#g')"
    local pattern3="$(echo "${package}" | sed 's#_#-#g')"

    if [ -n "${1}" ]
    then
        version="${1}"
        shift
    fi

    # if quiet operation is requested, redirect output
    if [ "${quiet}" = "yes" ] && [ -z "${outfile}" ]
    then
        outfile="/dev/null"
    fi

    # set include path for awk scripts
    export AWKPATH="${EISMAN_BIN_PATH}/awk"

    if [ -z "${outfile}" ]
    then
        ${gawk} -v incl_sections="${sections}"      \
                -v excl_sections="${exclsections}"  \
                -v packages="(((*|)${pattern1}(*|))|((*|)${pattern2}(*|))|((*|)${pattern3}(*|)))" \
                -v version="${version}"             \
                -v state="${state}"                 \
                -v url_filter="${filterurl}"        \
                -v fields="${fields}"               \
                -v database="${database}"           \
                -v is_tty="${is_tty}"               \
                -v color="${is_tty}"                \
                -f "${EISMAN_BIN_PATH}/awk/eisman_query.awk"
    else
        ${gawk} -v incl_sections="${sections}"      \
                -v excl_sections="${exclsections}"  \
                -v packages="(((*|)${pattern1}(*|))|((*|)${pattern2}(*|))|((*|)${pattern3}(*|)))" \
                -v version="${version}"             \
                -v state="${state}"                 \
                -v url_filter="${filterurl}"        \
                -v fields="${fields}"               \
                -v database="${database}"           \
                -v is_tty="no"                      \
                -v color="no"                       \
                -f "${EISMAN_BIN_PATH}/awk/eisman_query.awk" \
                > "${outfile}"
    fi

    exit ${?}
}

main "${@}"
