#!/bin/bash
# *******************************************************************
# **  @file eisman_uninstall.sh - uninstall package
# **
# **  Creation:     2013-01-13 dv
# **  Last update:  $Id: eisman_uninstall.sh 56090 2018-06-30 15:43:39Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

. /var/install/include/eislib
. /var/install/include/eismanlib

# list of protected system folders and files
system_files="\
(etc/init.d)|\
(etc/init.d/boot.d)|\
(etc/rc2.d)|\
(etc/check.d)|\
(etc/default.d)|\
(etc/profile.d)|\
(etc/bash_completion.d)|\
(etc/modprobe.d)|\
(etc/udev)|\
(etc)|\
(sbin)|\
(usr/share/doc)|\
(usr/share/locale)|\
(usr/share/man)|\
(usr/share/info)|\
(usr/share)|\
(usr/bin)|\
(usr/lib)|\
(usr/lib64)|\
(usr/sbin)|\
(usr/src)|\
(usr/include)|\
(usr/local/bin)|\
(usr/local/lib)|\
(usr/local/lib64)|\
(usr/local/sbin)|\
(usr/local/share)|\
(usr/local)|\
(usr)|\
(tmp)|\
(var/install)|\
(var/spool)|\
(var/lib)|\
(var)|\
(lib/modules)|\
(lib64)|\
(tmp/install.sh)|\
(tmp/preinstall.sh)"


# *******************************************************************
# **  show help on arguments and possible options
# **  @param  n/a
# **  @return n/a
# *******************************************************************

show_help()
{
cat <<EOF

Uninstall an eisfair package

usage:
    eisman uninstall [OPTIONS] PACKAGE

options:
    --help          show this help screen
    --force         uninstall even if requirements deny it
    --auto          non interactive mode
    --info          do not uninstall, only show packages
    --output=<file> redirect result into a file

EOF
}


# *******************************************************************
# **  remove package from database of installed files
# **  @param  $1  package name
# **  @return n/a
# *******************************************************************

remove_from_database()
{
    local package="${1}"

    # grep all packages other then "$package" into temporary file
    grep -v -e "^\[.*\]\[${package}\]\[.*\]\[.*\]\[.*\]" \
       "${EISMAN_PATH}/installed.db" > "${EISMAN_PATH}/installed.db.tmp"

    if [ "$?" -eq 0 ]
    then
        # remove database file
        rm -f "${EISMAN_PATH}/installed.db"

        # move temporary file to new database
        mv "${EISMAN_PATH}/installed.db.tmp" "${EISMAN_PATH}/installed.db"
    else
        echo "error: unable to delete package from database!" >&2
        exit 1
    fi
}


# *******************************************************************
# **  search installed package database for packages requireing the
# **  package that has to be uninstalled.
# **  @param  $1  package
# **  @return (on stdout) data found in packages.db meeting the
# **          search criteria
# *******************************************************************

find_backward_requirements()
{
    local package="${1}"

    result=$(grep -E "^\[.*\]\[.*\]\[.*\]\[.*\]\[.*\]require    ${package}([ ]|$)" \
        "${EISMAN_PATH}/installed.db")

    sed -ne 's#^\[.*\]\[\(.*\)\]\[\(.*\)\]\[.*\]\[.*\].*#- \1 (\2)#p' <<< "${result}"
}


# *******************************************************************
# **  return short description of the package
# **  @param  $1  package
# **  @return (on stdout) data found in packages.db meeting the
# **          search criteria
# *******************************************************************

get_package_description()
{
    local package="${1}"
    local fields="name|short|version|date"

    # perform query
    query_package_db      \
        ""                \
        ""                \
        "${package}"      \
        ""                \
        ""                \
        ""                \
        "${fields}"       \
        "${EISMAN_PATH}/installed.db"
}


# *******************************************************************
# **  pass result to logger
# **  @param  $1  package
# **  @param  $2  version
# **  @param  $3  message to be logged
# **  @param  $4  is it an update? (yes/no)
# **  @return n/a
# *******************************************************************

log_result()
{
    local package="${1}"
    local version="${2}"
    local message="${3}"
    local update="${4}"
    local date=$(sed -ne 's#^date  \(.*\)#\1#p' <<< "${description}" | sed 's# ##g')


    if [ "${update}" = "no" ]
    then
        /usr/bin/logger -t 'eis-deinstall' -p 'local5.info' \
            "package ${package} (${version}|${date}) ${message}"
    else
        /usr/bin/logger -t 'eis-update' -p 'local5.info' \
            "package ${package} (${version}|${date}) ${message}"
    fi
}


# *******************************************************************
# **  check if package is installed
# **  @param  $1  package
# **  @return 0 if installed, 1 otherwise
# *******************************************************************

is_package_installed()
{
    # check if package is installed
    local package=${1}
    local description=$(get_package_description "${package}")
    if [ -z "${description}" ]
    then
        return 1
    fi
    return 0
}


# *******************************************************************
# **  remove package entry from database and log result
# **  this routine is called to finalize uninstallation
# **  @param  $1  package
# **  @return 0 if successful
# *******************************************************************

uninstall_from_db()
{
    local package=${1}

    # only if this isn't an update, remove package record
    # from installed packages database
    if [ "${update}" = "no" ]
    then
        # remove database record and update database
        remove_from_database "${package}"
    fi

    # remove general kernel entry from database
    # if this entry still exists. Only the specialized
    # kernel package names are allowed
    if [ "${package:0:10}" = "eiskernel-" ]
    then
        remove_from_database "eiskernel"
    fi

    # success
    return 0
}


# *******************************************************************
# **  uninstall a package using the filelist supplied by the package
# **  (if file exists).
# **  @param  $1  package
# **  @param  $2  argument (may optionally be "--update")
# **  @return 0 if successful
# *******************************************************************

uninstall_files()
{
    local package=${1}
    local argument=${2}
    local file_filter="${system_files}"

    # if this is an update, make sure no config file is going
    # to be deleted
    if [ "${argument}" = "--update" ]
    then
        file_filter="${file_filter}|(etc/config.d/.*)"
    fi

    # check if file list exists, exit if not
    filelist=/etc/filelist.d/${package}-files.txt
    if [ ! -f ${filelist} ]
    then
        return 0
    fi

    # - sort file list in reverse order,
    # - remove trailing / from file and path names
    # - filter comment lines
    # - filter protected system files and folders
    # - write to temporary file
    sort -r -b -k 6 ${filelist}         |
        sed 's#/*$##'                    |
        egrep -v "^\ *#.*$"              |
        egrep -v "^.* (${file_filter})$" \
        > /tmp/tmpfile$$

    # process file list info and remove files if temporary file
    # has been created and is not empty
    if [ -s /tmp/tmpfile$$ ]
    then
        while read id perm user group packname file
        do
            case ${id} in
            b|u|s|h)
                # remove program files
                rm -f /${file}
                ;;
            f)
                # remove directories
                rmdir --ignore-fail-on-non-empty /${file} 2>/dev/null
                ;;
            esac
        done < /tmp/tmpfile$$
    fi

    # make sure the file list is removed too
    rm -f /tmp/tmpfile$$
    rm -f ${filelist}

    return 0
}


# *******************************************************************
# **  run package uninstall script (if present). Exit with error
# **  if script fails
# **  @param  $1  package
# **  @param  $2  argument (''|'--update')
# **  @return 0 if successful, 1 otherwise
# *******************************************************************

run_uninstall_script()
{
    local package=${1}
    local argument=${2}

    # check if uninstall script exists, exit if not
    if [ ! -f /var/install/deinstall/${package} ]
    then
        return 0
    fi

    # modify file permissions
    chmod 0750 /var/install/deinstall/${package}

    # convert the file to unix :-(
    /usr/bin/dtou /var/install/deinstall/${package}

    # execute deinstallation and verify result
    /var/install/deinstall/${package} ${argument}
    if [ ${?} -ne 0 ]
    then
        # log result and exit with error code 1
        return ${?}
    fi

    # make sure uninstall script is removed too
    rm -f /var/install/packages/${package}
    return 0
}


# *******************************************************************
# **  uninstall a package by checking for proper uninstall method
# **  @param  $1  package
# **  @return 0 if successful, 1 otherwise
# *******************************************************************

uninstall_package()
{
    local package=${1}
    local version

    # check if package is currently installed and get package info
    description=$(get_package_description "${package}")
    if [ -z "${description}" -a "${force}" = "no" ]
    then
        echo "error: package \"${package}\" not installed!" >&2
        return 1
    fi

    # retreive package version
    version=$(sed -ne 's#^version  \(.*\)#\1#p' <<< "${description}" | sed 's# ##g')

    # prepare argument for update
    argument=""
    if [ "${update}" = "yes" ]
    then
        argument="--update"
    fi

    # check if uninstallation is possible
    if [ -f /var/install/deinstall/${package}  \
      -o -f /etc/filelist.d/${package}-files.txt ]
    then
        # continue with normal uninstall procedure by
        # 1. calling package uninstall script (if present)
        # 2. uninstalling files from provided file list
        # 3. remove entry from package database
        if run_uninstall_script ${package} ${argument} && \
           uninstall_files      ${package} ${argument} && \
           uninstall_from_db    ${package}
        then
            # print result
            if [ "${update}" = "no" ] && [ "${quiet}" = "no" ]
            then
                mecho --info -n "Successfully uninstalled: "
                mecho        -n "${package} (${version})"
                mecho --info    "!"
            fi

            # log result and exit with success code 0
            # this procedure will never fail.. hopefully ;-)
            log_result "${package}" "${version}" \
                       "without error" "${update}"
            return 0
        else
            local rc=${?}

            # print result
            mecho --warn -n "Failed to uninstall: "
            mecho        -n "${package} (${version})"
            mecho --info    "!"

            # log result and exit with error code 1
            log_result "${package}" "${version}" \
                       "with error(s) (${rc}/0)" "${update}"
            return 1
        fi
    else
        if [ "${force}" = "yes" ] || [ "${update}" = "yes" ]
        then
            # log result and exit with success code 0
            # this procedure will never fail.. hopefully ;-)
            log_result "${package}" "${version}" \
                    "missing files, uninstall forced!" "${update}"

            # show warning message only if we are not updating
            if [ "${update}" = "no" ]
            then
                mecho --warn -n "Missing files, uninstall forced: "
                mecho           "${package} (${version})!"
            fi

            # only if this isn't an update, remove package record
            # from installed packages database
            if [ "${update}" = "no" ]
            then
                # remove database record and update database
                remove_from_database "${package}"
            fi

            return 0
        else
            # missing either file list or deinstall script
            echo "Unable to deinstall package ${package}!"
        fi
    fi

    # exit with error
    return 1
}


# *******************************************************************
# **  main program
# *******************************************************************

main()
{
    local outfile=""

    force='no'
    auto='no'
    update='no'
    quiet='no'
    info='no'

    while [ "${1:0:2}" = "--" ]
    do
        case ${1} in
        --help)
            show_help
            exit 0
            ;;
        --force)
            force='yes'
            shift
            ;;
        --auto)
            shift
            auto='yes'        # non interactive installation
            ;;
        --info)
            shift
            info='yes'        # not a real uninstallation only show what would be done
            ;;
        --quiet)
            shift
            quiet='yes'       # no success messages
            ;;
        --update)
            shift
            update='yes'
            ;;
        --output)
            shift
            outfile="${1}"
            shift
            ;;
        --output=*)
            outfile="${1:9}"
            shift
            ;;
        *)
            echo "error: unknown option \"${1}\"!" >&2
            echo
            show_help
            exit 1
            ;;
        esac
    done

    # read package parameter
    package="${1}"
    if [ -z "${package}" ]
    then
        echo "error: missing package parameter!" >&2
        echo
        show_help
        exit 1
    fi


    # creation of package list

    # if package is not installed, exit if force is not set
    if ! is_package_installed "${package}"
    then
        if [ "${quiet}" = "no" ]
        then
            echo "info: package \"${package}\" not installed!"
        fi

        if [ "${force}" = "no" ]
        then
            exit 0
        fi
    fi

    # start creation of package list
    packagelist[0]="${package}"
    num_packages=1

    if [ "${update}" = "no" ]
    then
        # if we are uninstalling a library, remove -dev package too
        if is_package_installed "${package}-dev"
        then
            packagelist[${num_packages}]="${package}-dev"
            num_packages=$((${num_packages} + 1))
        fi

        # if we are uninstalling a library, remove -dev-static package too
        if is_package_installed "${package}-dev-static"
        then
            packagelist[${num_packages}]="${package}-dev-static"
            num_packages=$((${num_packages} + 1))
        fi

        # if we only uninstall a lib-dev package, the lib-dev-static
        # package must be uninstalled to
        if is_package_installed "${package}-static"
        then
            packagelist[${num_packages}]="${package}-static"
            num_packages=$((${num_packages} + 1))
        fi
    fi


    # check if the package may be uninstalled
    if [ "${force}" = "no" -a "${update}" = "no" ]
    then
        index=0
        while [ "${index}" -lt "${num_packages}" ]
        do
            requirements=$(find_backward_requirements "${packagelist[${index}]}")

            for requirement in ${requirements}
            do
                if [ "${requirement}" != '-' ] && [ "${requirement:0:1}" != '(' ]
                then
                    index2=0
                    while [ "${index2}" -lt "${num_packages}" ]
                    do
                        if [ "${packagelist[${index2}]}" = "${requirement}" ]
                        then
                            break
                        fi
                        index2=$((${index2} + 1))
                    done

                    if [ "${index2}" -ge "${num_packages}" ]
                    then
                        echo "error: uninstallation aborted due to pending requirements!" >&2

                        mecho --info "List of pending requirements:"
                        ${EISMAN_BIN_PATH}/eisman_usage.sh "${package}"

                        exit 1
                    fi
                fi
            done

            index=$((${index} + 1))
        done
    fi


    # ask user to continue
    if [ "${update}" = "no" ] && [ "${auto}" = "no" ]
    then
        # check if package is currently installed and get package info
        description=$(get_package_description "${package}")

        if [ -z "${outfile}" ]
        then
            # show info
            mecho --info 'The following package will be uninstalled:'
            echo "${description}"
            echo

            if [ "${num_packages}" -gt 1 ]
            then
                mecho --info 'This will also trigger uninstallation of:'
                index=1
                while [ "${index}" -lt "${num_packages}" ]
                do
                    echo "- ${packagelist[${index}]}"
                    index=$((${index} + 1))
                done
                echo
            fi
        else
            {
                # show info
                mecho --info 'The following package will be uninstalled:'
                echo "${description}"
                echo

                if [ "${num_packages}" -gt 1 ]
                then
                    mecho --info 'This will also trigger uninstallation of:'
                    index=1
                    while [ "${index}" -lt "${num_packages}" ]
                    do
                        echo "- ${packagelist[${index}]}"
                        index=$((${index} + 1))
                    done
                    echo
                fi
            } > "${outfile}"
        fi

        if [ "${info}" = "no" ]
        then
            res=$(/var/install/bin/ask 'Continue' 'no')

            # if ask break, ask returned 255
            if [ $? = 255 ]
            then
                res='no'
            fi

            if [ "${res}" != "yes" ]
            then
                echo 'Uninstall aborted by user!' >&2
                exit 1
            fi
        fi
    fi

    if [ "${info}" = "no" ]
    then
        index=0
        result=0
        while [ "${index}" -lt "${num_packages}" ]
        do
            if ! uninstall_package "${packagelist[${index}]}"
            then
                mecho --warn "Package ${packagename} NOT removed."
                result=1
            fi
            index=$((${index} + 1))
        done
    fi

    exit $result
}

main "${@}"
