#!/bin/sh
# *******************************************************************
# **  @file eisman_upgrade.sh - upgrade all stable packages
# **
# **  Creation:     2016-05-27 dv
# **  Last update:  $Id: eisman_upgrade.sh 65500 2019-05-25 07:08:28Z dv $
# **
# **  Copyright (c) 2013-2019 the eisfair team, team(at)eisfair(dot)org
# **
# **  This program is free software; you can redistribute it and/or modify
# **  it under the terms of the GNU General Public License as published by
# **  the Free Software Foundation; either version 2 of the License, or
# **  (at your option) any later version.
# *******************************************************************

. /var/install/include/eislib
. /var/install/include/eismanlib

gawk=/usr/bin/gawk

# *******************************************************************
# **  show help on arguments and possible options
# **  @param  n/a
# **  @return n/a
# *******************************************************************

show_help()
{
cat <<EOF

Update packages of a system

usage:
    eisman upgrade [OPTIONS]

options:
    --unstable          install unstable packages too
    --listonly          only produce output like eisman queury does
    --no-wait-anim      do not show wait animation
    --output=<file>     redirect result into a file. This is only valid
                        if --listonly is specified too
    --help              show this help screen
    --exclude=<package> don't add package to the list of upgradable
                        packages. May be specified multiple times.
                        Note: This will not prevent the package from
                        being installed if it is required by some other
                        package!
    --skip=<package>    Same as --exclude
    --no-tty            don't use escape sequences for cursor movement
                        and color output.

EOF
}


# *******************************************************************
# **  main routine
# **  @param  $@ list of command line arguments
# **  @return n/a
# *******************************************************************

main()
{
    local unstable="no"
    local action="install"
    local outfile
    local arguments=""
    local wait_animation="yes"
    local skiplist=""

    # check if output device is a pipe
    if [ -t 1 ]
    then
        local is_tty="yes"
        local action="install"
    else
        local is_tty="no"
        local action="list"
    fi

    while [ "${1:0:2}" = "--" ]
    do
        case $1 in
        --help)
            show_help
            return 0
            ;;
        --listonly|--info)
            shift
            action="list"
            ;;
        --unstable)
            shift
            unstable="yes"
            arguments="${arguments} --unstable"
            ;;
        --no-wait-anim)
            wait_animation="no"
            arguments="${arguments} --no-wait-anim"
            shift
            ;;
        --output)
            shift
            outfile="${1}"
            shift
            ;;
        --output=*)
            outfile="${1:9}"
            shift
            ;;
        --skip|--exclude)
            shift
            if [ -n "${skiplist}" ]
            then
                skiplist="${skiplist},${1}"
            else
                skiplist="${1}"
            fi
            shift
            ;;
        --exclude=*)
            if [ -n "${skiplist}" ]
            then
                skiplist="${skiplist},${1:10}"
            else
                skiplist="${1:10}"
            fi
            shift
            ;;
        --skip=*)
            if [ -n "${skiplist}" ]
            then
                skiplist="${skiplist},${1:7}"
            else
                skiplist="${1:7}"
            fi
            shift
            ;;
        --no-tty)
            is_tty="no"
            shift
            ;;
        *)
            echo "error: unknown option \"${1}\"!" >&2
            echo ""
            show_help
            return 1
            ;;
        esac
    done

    # if an output file has been specified, action is list
    if [ -n "${outfile}" ]
    then
        action="list"
        is_tty="no"
    fi

    # set include path for awk scripts
    export AWKPATH="${EISMAN_BIN_PATH}/awk"

    if [ -z "${outfile}" ]
    then
        ${gawk} -v unstable="${unstable}"           \
                -v action="${action}"               \
                -v exclude="${skiplist}"            \
                -v progress="${wait_animation}"     \
                -v is_tty="${is_tty}"               \
                -v color="${is_tty}"                \
                -f "${EISMAN_BIN_PATH}/awk/eisman_upgrade.awk"
    else
        ${gawk} -v unstable="${unstable}"           \
                -v action="${action}"               \
                -v exclude="${skiplist}"            \
                -v progress="${wait_animation}"     \
                -v is_tty="${is_tty}"               \
                -v color="${is_tty}"                \
                -f "${EISMAN_BIN_PATH}/awk/eisman_upgrade.awk" \
                > "${outfile}"
    fi

    exit ${?}
}

main "${@}"
